local famBags = LibStub('AceAddon-3.0'):GetAddon('famBags')
famBags.Sections = {
   bags = {},
   bank = {}
}
famBags.Sections.Bin = {
   bags = {},
   bank = {}
}
-- :::::::::::: LOCALS ::::::::::::: -------------------------------------------------------------------------

-- :::::::::::: famBags.Sections ::::::::::::: -------------------------------------------------------------------------
function famBags.Sections:Create(sectionType, bag, index, title)
   local db = famBags.db.profile
   self[bag][sectionType] = self[bag][sectionType] or {}
   
   if not self[bag][sectionType][index] then
      local sectionIndex = (sectionType == 'bags') and 'bags' or index
      db.sections[sectionIndex] = db.sections[sectionIndex] or {}
      db.sections[sectionIndex].bg = db.sections[sectionIndex].bg or famBags:GetBGDefaults()
      db.sections[sectionIndex].title = db.sections[sectionIndex].title or famBags:GetTitleDefaults()
      
      self[bag][sectionType][index] = famBags:CreateBagFrame(
         string.format('famBags_%s_%s_%s', bag, sectionType, index), 
         title, (bag == 'bags') and famBags.famBagsFrame or famBags.famBankFrame, 
         2, 100, 100, db.sections[sectionIndex])
         
      self[bag][sectionType][index].sectionType = sectionType
      self[bag][sectionType][index].name = index
   end
   
   return self[bag][sectionType][index]
end

function famBags.Sections:Destroy(sectionType, index)
   for bag in pairs(famBags.BagIndexes) do
      if self[bag] and self[bag][sectionType] and self[bag][sectionType][index] then
         self[bag][sectionType][index]:Hide()
         self[bag][sectionType][index] = nil
      end
   end
end


function famBags.Sections:HideSectionsType(bag, sectionType)
   if self[bag][sectionType] then
      for _, bag in pairs(self[bag][sectionType]) do
         bag:Hide()
      end
   end
end

function famBags.Sections:HideSection(bag, sectionType, index, layoutBags)
   if self[bag][sectionType][index] then
      self[bag][sectionType][index]:Hide()
--[[      if layoutBags then
         self:LayoutBags(bag, sectionType)
      end]]   
   end
end

function famBags.Sections:SetTitle(bag, section, title)
   local db = (bag == 'bags') and famBags.db.profile.sections.bags.title.text or famBags.db.profile.sections[section.name].title.text
   if section.name == BANK_CONTAINER then
      title = famBags.L['Bank Main Container']
   end
   famBags:SetTextLayout(section.titleFrame, 'titleText', db, title)   
end

function famBags.Sections:BuildBagsSections(bag)
   for _, i in ipairs(famBags.BagIndexes[bag]) do
      local bagname = GetBagName(i)
      self:Create('bags', bag, i, bagname)
      if not bagname then
         self:HideSection(bag, 'bags', i)
      end
   end
end

function famBags.Sections:BuildCategoriesSections(bag)
   local db = famBags.db.profile.categories
   self[bag].categories = self[bag].categories or {}
   self[bag].categories = self[bag].categories or {}
   for _, cat in pairs(famBags.SortedCategories) do
      if not self[bag].categories[cat] then
         self:Create('categories', bag, cat, strsub(cat, 4))
      end
      self[bag].categories[cat]:Hide()
   end
end

function famBags.Sections:WipeBin(bag)
   for _, row in ipairs(self.Bin[bag]) do
      if row.cells then wipe(row.cells) end
      row.width = 0
   end
end

function famBags.Sections:PushToBin(bag, section)
   local db = famBags.db.profile.layout
   local bw = (famBags.Buttons[0] and famBags.Buttons[0][1]) and famBags.Buttons[0][1]:GetWidth() or famBags.Buttons[BANK_CONTAINER][1]:GetWidth()
   local maxwidth = ((bw * db.general.buttons_x_row) + (db.buttons.hspacing * (db.general.buttons_x_row - 1)) + (db.buttons.hpadding * 2)) * db.buttons.size
   local hs = db.sections.hspacing
   local located
   local width = section:GetWidth()
   for i, row in ipairs(self.Bin[bag]) do
      if row.width == 0 or row.width + hs + width <= maxwidth then
         row.width = row.width + hs + width
         row.cells[#row.cells + 1] = section
         return
      end
   end
   local i = #self.Bin[bag] + 1
   self.Bin[bag][i] = {}
   self.Bin[bag][i].width = width
   self.Bin[bag][i].cells = { section }
end

function famBags.Sections:UpdateBin(bag, sectionType)
   self:WipeBin(bag)
   local sections = (sectionType == 'bags') and self[bag]['bags'] or famBags.SortedCategories
   for k, section in pairs(sections) do
      section = (sectionType == 'categories') and self[bag]['categories'][section] or section
      if section:IsVisible() then
         self:PushToBin(bag, section)
      end
   end
end

function famBags.Sections:LayoutBags(bag, sectionType, clear)
   if clear then
      if sectionType ~= 'bags' then
         self:HideSectionsType(bag, 'bags')
      end
      if sectionType ~= 'categories' then
         self:HideSectionsType(bag, 'categories')
      end
   end
   if self[bag][sectionType] then
      local frame = (bag == 'bags') and famBags.famBagsFrame or famBags.famBankFrame
      local dbbag = (bag == 'bags') and 'mainwindow' or 'bank'
      local db = famBags.db.profile.layout.sections
      local previous = frame.bodyFrame
      local th = frame.titleFrame:GetHeight()
      local h = db.vpadding + th
      local lw = 0
      local lh = 0
      local w = 0
      local y = db.vpadding
      local hsection
      if famBags.db.profile[dbbag].title.is_inset then
         y = th + (famBags.db.profile[dbbag].title.inset * 2)
         h = h + (famBags.db.profile[dbbag].title.inset * 2)
      end
      y = -y
      self:UpdateBin(bag, sectionType)
      local bin = self.Bin[bag]
      for i, row in ipairs(bin) do
         lw, lh = 0, 0
         if row.width == 0 then break end
         for j, section in pairs(row.cells) do
            section:ClearAllPoints()
            if j == 1 then
               if i == 1 then
                  section:SetPoint('TOPLEFT', frame.bodyFrame, 'TOPLEFT', db.hpadding, y)
               else
                  section:SetPoint('LEFT', bin[i-1].cells[1], 'LEFT', 0, 0)
                  section:SetPoint('TOP', hsection, 'BOTTOM', 0, -db.vspacing)
               end
               lh = section:GetHeight()
               hsection = section
            else
               section:SetPoint('TOPLEFT', previous, 'TOPRIGHT', db.hspacing, 0)
               lw = lw + db.hspacing
               local alh = section:GetHeight()
               if alh > lh then
                  lh = alh
                  hsection = section
               end
            end
            lw = lw + section:GetWidth()
            previous = section
            famBags:SetTextLayout(section.titleFrame, 'titleText', famBags.db.profile.sections[section.name].title.text, nil)
         end
         h = h + db.vspacing + lh
         w = max(w, lw)
      end
      h = h - db.vspacing + db.vpadding
      w = w + (db.hpadding * 2)
      if bag == 'bank' and famBags.db.profile.bankwindow.enabled and famBags.famBagsBlizzardBankFrame then
         local bbw = famBags.famBagsBlizzardBankFrame.PurchaseInfo:GetWidth()
         if w < bbw then
            w = bbw
         end
      end
      frame:SetWidth(w)
      frame:SetHeight(h)
   end
end
