local famBagsOptions = LibStub('AceAddon-3.0'):NewAddon('famBagsOptions', 'AceEvent-3.0')
local famBags = LibStub('AceAddon-3.0'):GetAddon('famBags')

-- :::::::::::: LOCALS ::::::::::::: -------------------------------------------------------------------------

local L = famBags.L

local LSM = LibStub('LibSharedMedia-3.0', true)
LoadAddOn('AceGUI-3.0-SharedMediaWidgets')
local AceGUI = LibStub('AceGUI-3.0')

local bg_types = {
   L['Solid'],
   L['Vertical gradient'],
   L['Horizontal gradient'],
   L['Texture']
}
local text_flags = {
   ['OUTLINE'] = L['OutLine'],
   ['THICKOUTLINE'] = L['Thick outline'],
   ['MONOCHROME'] = L['Monochrome'],
   ['SHADOW'] = L['Shadow'],
}
local horizontal_align = {
   ['LEFT'] = L['Left'],
   ['RIGHT'] = L['Right'],
   ['CENTER'] = L['Center'],
}
local vertical_align = {
   ['TOP'] = L['Top'],
   ['BOTTOM'] = L['Bottom'],
   ['CENTER'] = L['Center'],
}
local logical_operators = {
   ['AND'] = L['AND'],
   ['AND NOT'] = L['AND NOT'],
   ['OR'] = L['OR'],
   ['OR NOT'] = L['OR NOT'],
}

local quality_glows = {
   ['none'] = L['None'],
   ['square'] = L['Square'],
   ['round1'] = L['Light Rounded'],
   ['round2'] = L['Heavy Rounded'],
   ['circle'] = L['Circle'],
}

local CURRENT_SECTION
local CURRENT_SECTION_DB
local SECTION_ARGS
local SECTIONS = {
   ['bags'] = L['Bags'],
}
local SECTIONS_TYPES = {
   ['onebag'] = L['One Bag'],
   ['bags'] = L['Bags'],
   ['categories'] = L['Categories'],
}
local CATEGORIES_OPTIONS

local function get_array_len(arr)
   local i = 0
   for _ in pairs(arr) do
      i = i + 1
   end
   return i
end

local function get_max_order(arr)
   local i = 0
   for _, opt in pairs(arr) do
      if opt and opt.order then
         i = max(i, opt.order)
      end
   end
   return i
end

local function get_actual_category_name(category)
   local db = famBags.db.profile.categories
   category = strsub(category, 4)
   for n in pairs(db) do
      if strsub(n, 4) == category then
         return n
      end
   end
   return category
end

local function select_category(cat)
   LibStub('AceConfigDialog-3.0'):SelectGroup('famBags Categories', cat)
end

local function get_types()
   local types = {
      [' '] = ' '
   }
   for k in pairs(famBags.Types) do
      types[k] = k
   end
   return types
end

local function get_subtypes(t, set)
   local subtypes = {
      [L['.: All :.']] = L['.: All :.'],
   }
   if t then
      for _, k in pairs(famBags[set][t]) do
         subtypes[k] = k
      end
   end
   return subtypes
end

local function get_criteria_index(cat, t, set)
   cat = get_actual_category_name(cat)
   set = set or 'Types'
   if set == 'Types' or type(t) == 'number' then return t end
   local db = famBags.db.profile.categories[cat][set]
   for i, c in ipairs(db) do
      if c.name == t then
         return i
      end
   end
   return nil
end

local function get_type_db(cat, t, set)
   if not famBags.db.profile.categories[cat] then return nil end
   local db = famBags.db.profile.categories[cat]
   db[set] = db[set] or {}
   db = db[set]
   if t then
      if set == 'Criterias' then
         db = db[get_criteria_index(cat, t, set)].criterias
      else
         db = db[t]
      end
   end
   return db
end

local function delete_type(cat, t, set)
   cat = get_actual_category_name(cat)
   local db = get_type_db(cat, nil, set)
   local i = get_criteria_index(cat, t, set)
   if db and db[i] then
      local args = (set == 'Types') and CATEGORIES_OPTIONS.args[cat].args.types_set.args or CATEGORIES_OPTIONS.args[cat].args.criterias_set.args
      if set == 'Types' then
         db[i] = nil
      else
         table.remove(db, i)
      end
      args[t..'_oper'] = nil
      args[t] = nil
      args[t..'_up'] = nil
      args[t..'_down'] = nil
      args[t..'_del'] = nil
      args[t..'_sep'] = nil
      LibStub('AceConfigRegistry-3.0'):NotifyChange('famBags Categories')
      select_category(cat)
   end
end

local function add_criteria(cat, c)
   cat = get_actual_category_name(cat)
   local db = famBags.db.profile.categories[cat]
   db.Criterias = db.Criterias or {}
   if db.Criterias[c] then return end
   local newcrit = {
      name = c,
      criterias = {
         [L['.: All :.']] = L['.: All :.']
      },
      operator = 'AND'
   } 
   table.insert(db.Criterias, newcrit)
   local cs = CATEGORIES_OPTIONS.args[cat].args.criterias_set.args
   local order = get_max_order(cs) + 1
   cs[c..'_oper'] = famBagsOptions:GetCategoryOperatorOptions(cat, c, order)
   cs[c] = famBagsOptions:GetCategoryTypeOptions(cat, c, order + 1, 'Criterias')
   cs[c..'_up'] = famBagsOptions:GetCategoryTypeUp(cat, c, order + 2)
   cs[c..'_down'] = famBagsOptions:GetCategoryTypeDown(cat, c, order + 3)
   cs[c..'_del'] = famBagsOptions:GetCategoryTypeDelete(cat, c, order + 4, 'Criterias')
   cs[c..'_sep'] = famBagsOptions:GetCategoryTypeSeparator(order + 5, 'Criterias')
   LibStub('AceConfigRegistry-3.0'):NotifyChange('famBags Categories')
   select_category(cat)
end

local function add_type(cat, t)
   cat = get_actual_category_name(cat)
   local db = famBags.db.profile.categories[cat]
   db.Types = db.Types or {}
   if db.Types[t] then return end
   db.Types[t] = {}
   db.Types[t][L['.: All :.']] = L['.: All :.']
   local ts = CATEGORIES_OPTIONS.args[cat].args.types_set.args
   local order = get_max_order(ts) + 1
   ts[t] = famBagsOptions:GetCategoryTypeOptions(cat, t, order)
   ts[t..'_del'] = famBagsOptions:GetCategoryTypeDelete(cat, t, order + 1)
   ts[t..'_sep'] = famBagsOptions:GetCategoryTypeSeparator(order + 2)
   LibStub('AceConfigRegistry-3.0'):NotifyChange('famBags Categories')
   select_category(cat)
end

local function get_criterias()
   local criterias = {
      [' '] = ' '
   }
   for k in pairs(famBags.Criterias) do
      criterias[k] = k
   end
   return criterias
end

local function move_criteria(cat, t, step)
   cat = get_actual_category_name(cat)
   local db = get_type_db(cat, nil, 'Criterias')
   local i = get_criteria_index(cat, t, 'Criterias')
   if not (db and db[i + step]) then return end
   local tstep = db[i + step].name
   db[i], db[i + step] = db[i + step], db[i]
   local args = (set == 'Types') and CATEGORIES_OPTIONS.args[cat].args.types_set.args or CATEGORIES_OPTIONS.args[cat].args.criterias_set.args
   args[t..'_oper'].order, args[tstep..'_oper'].order = args[tstep..'_oper'].order, args[t..'_oper'].order
   args[t].order, args[tstep].order = args[tstep].order, args[t].order
   args[t..'_up'].order, args[tstep..'_up'].order = args[tstep..'_up'].order, args[t..'_up'].order
   args[t..'_down'].order, args[tstep..'_down'].order = args[tstep..'_down'].order, args[t..'_down'].order
   args[t..'_del'].order, args[tstep..'_del'].order = args[tstep..'_del'].order, args[t..'_del'].order
   args[t..'_sep'].order, args[tstep..'_sep'].order = args[tstep..'_sep'].order, args[t..'_sep'].order
end

local function get_category_by_index(index)
   local db = famBags.db.profile.categories
   for name in pairs(db) do
      local i = tonumber(strsub(name, 1, 2))
      if (i == index) then
         return name
      end
   end
   return nil
end

local function change_category_key(old, new)
   local db = famBags.db.profile.categories
   local c = CATEGORIES_OPTIONS.args
   db[old], db[new] = nil, db[old]
   c[old], c[new] = nil, c[old]
   c[new].name = new
   db = famBags.db.profile.sections
   db[old], db[new] = nil, db[old]
   SECTIONS[old], SECTIONS[new] = nil, new
   for bag in pairs(famBags.BagIndexes) do
      if famBags.Sections[bag] and famBags.Sections[bag].categories and famBags.Sections[bag].categories[old] then
         famBags.Sections[bag].categories[old], famBags.Sections[bag].categories[new] = nil, famBags.Sections[bag].categories[old]
         famBags.Sections[bag].categories[new].name = new
      end
      if famBags.db.profile.layout.general.sections == 'categories' and 
         famBags.Buttons.Current[bag] and famBags.Buttons.Current[bag][old] then
         famBags.Buttons.Current[bag][old], famBags.Buttons.Current[bag][new] = nil, famBags.Buttons.Current[bag][old]
      end
   end
end

local function delete_category(category)
   local db = famBags.db.profile.categories
   local c = CATEGORIES_OPTIONS.args
   category = get_actual_category_name(category)
   db[category] = nil
   c[category] = nil
   famBags.Sections:Destroy('categories', category)
   SECTIONS[category] = nil
   local ri = tonumber(strsub(category, 1, 2))
   local t = {}
   for name in pairs(db) do
      local ci = tonumber(strsub(name, 1, 2))
      if (ci > ri) then
         local key = string.format('%02d', (ci - 1))..'.'..strsub(name, 4)
         t[name] = key
      end
   end
   for old, new in pairs(t) do
      change_category_key(old, new)
   end
   famBags:BuildSortedCategories()
   LibStub('AceConfigRegistry-3.0'):NotifyChange('famBags Categories')
   LibStub('AceConfigRegistry-3.0'):NotifyChange('famBags Sections')
end

local function move_category_up(cat)
   cat = get_actual_category_name(cat)
   local ri = tonumber(strsub(cat, 1, 2))
   if ri < 2 then return end
   local swap = get_category_by_index(ri - 1)
   local newcat, newswap = strsub(swap, 1, 2)..'.'..strsub(cat, 4), strsub(cat, 1, 2)..'.'..strsub(swap, 4)

   change_category_key(cat, newcat)
   change_category_key(swap, newswap)
   famBags:BuildSortedCategories()

   LibStub('AceConfigRegistry-3.0'):NotifyChange('famBags Categories')
   LibStub('AceConfigRegistry-3.0'):NotifyChange('famBags Sections')

   select_category(newcat)
end

local function move_category_down(cat)
   cat = get_actual_category_name(cat)
   local ri = tonumber(strsub(cat, 1, 2))
   if ri >= get_array_len(famBags.db.profile.categories) then return end
   local swap = get_category_by_index(ri + 1)
   local newcat, newswap = strsub(swap, 1, 2)..'.'..strsub(cat, 4), strsub(cat, 1, 2)..'.'..strsub(swap, 4)

   change_category_key(cat, newcat)
   change_category_key(swap, newswap)
   famBags:BuildSortedCategories()

   LibStub('AceConfigRegistry-3.0'):NotifyChange('famBags Categories')
   LibStub('AceConfigRegistry-3.0'):NotifyChange('famBags Sections')

   select_category(newcat)
end

local function copy_section_db(from)
   famBags.db.profile.sections[CURRENT_SECTION] = CopyTable(famBags.db.profile.sections[from])
   CURRENT_SECTION_DB = famBags.db.profile.sections[CURRENT_SECTION]
end

local function copy_data(from, to)
   for k, v in pairs(from) do
      if type(v) == 'table' then
         to[k] = to[k] or {}
         copy_data(from[k], to[k])
      else
         to[k] = v
      end
   end
end

-- :::::::::::: famBagsOptions ::::::::::::: -------------------------------------------------------------------------
function famBagsOptions:OnInitialize()
   self.db = famBags.db
   local acr = LibStub('AceConfigRegistry-3.0')
   acr:RegisterOptionsTable('famBags', self:GetGlobalOptions())
   acr:RegisterOptionsTable('famBags Main bags', self:GetMainWindowOptions())
   acr:RegisterOptionsTable('famBags Bank', self:GetBankOptions())
   acr:RegisterOptionsTable('famBags Money', self:GetMoneyOptions())
   acr:RegisterOptionsTable('famBags Blizzard bank', self:GetBankWindowOptions())
   acr:RegisterOptionsTable('famBags Bags slots', self:GetBagsWindowOptions())
   acr:RegisterOptionsTable('famBags Sections', self:GetSectionsOptions())
   acr:RegisterOptionsTable('famBags Categories', self:GetCategoriesOptions())
   acr:RegisterOptionsTable('famBags Presets', self:GetPresetsOptions())
   acr:RegisterOptionsTable('famBags Profiles', LibStub("AceDBOptions-3.0"):GetOptionsTable(famBags.db))

   local acd = LibStub('AceConfigDialog-3.0')
   acd:AddToBlizOptions('famBags', 'famBags')
   acd:AddToBlizOptions('famBags Main bags', 'Main bags', 'famBags')
   acd:AddToBlizOptions('famBags Bank', 'Bank', 'famBags')
   acd:AddToBlizOptions('famBags Money', 'Money', 'famBags')
   acd:AddToBlizOptions('famBags Blizzard bank', 'Blizzard bank', 'famBags')
   acd:AddToBlizOptions('famBags Bags slots', 'Bags slots', 'famBags')
   acd:AddToBlizOptions('famBags Sections', 'Sections', 'famBags')
   acd:AddToBlizOptions('famBags Categories', 'Categories', 'famBags')
   acd:AddToBlizOptions('famBags Presets', 'Presets', 'famBags')
   acd:AddToBlizOptions('famBags Profiles', 'Profiles', 'famBags')
   
  famBags.db.RegisterCallback(self, "OnProfileChanged", "FullRefresh")
  famBags.db.RegisterCallback(self, "OnProfileCopied", "FullRefresh")
  famBags.db.RegisterCallback(self, "OnProfileReset", "FullRefresh")   
end

function famBagsOptions:OnEnable()
   --self:RegisterEvent('PLAYER_ENTERING_WORLD')TOTO =
end

function famBagsOptions:FullRefresh()
   famBags:FullRefresh()
   CURRENT_SECTION = nil
   CURRENT_SECTION_DB = nil
   SECTION_ARGS = nil
   SECTIONS = {
      ['bags'] = L['Bags'],
   }
   local CATEGORIES_OPTIONS = nil
   local acr = LibStub('AceConfigRegistry-3.0')
   acr:RegisterOptionsTable('famBags', self:GetGlobalOptions())
   acr:RegisterOptionsTable('famBags Main bags', self:GetMainWindowOptions())
   acr:RegisterOptionsTable('famBags Bank', self:GetBankOptions())
   acr:RegisterOptionsTable('famBags Money', self:GetMoneyOptions())
   acr:RegisterOptionsTable('famBags Blizzard bank', self:GetBankWindowOptions())
   acr:RegisterOptionsTable('famBags Bags slots', self:GetBagsWindowOptions())
   acr:RegisterOptionsTable('famBags Sections', self:GetSectionsOptions())
   acr:RegisterOptionsTable('famBags Categories', self:GetCategoriesOptions())
end

function famBagsOptions:GetGlobalOptions()
   local db = self.db.profile.layout
   local function get_opt(info)
      local id = info[#info]
      local section = info[#info-1]
      return db[section][id]
   end
   local function set_opt(info, value)
      local id = info[#info]
      local section = info[#info-1]
      db[section][id] = value
      if section..id == 'generalsections' then
         famBags.Sections:LayoutBags('bags', db.general.sections, true)
         famBags.Sections:LayoutBags('bank', db.general.sections, true)
      end
      famBags:LayoutBagsButtons()
   end
   return {
      name = L['famBags'],
      type = 'group',
      args = {
         general = {
            name = L['Global layout'],
            type = 'group',
            inline = true,
            args = {
               sections = {
                  type = 'select',
                  name = L['Layout'],
                  get = get_opt,
                  set = set_opt,
                  values = function(info)
                     return SECTIONS_TYPES
                  end,
                  order = 1,
               },
               ordered = {
                  type = 'toggle',
                  name = L['Order items'],
                  get = get_opt,
                  set = set_opt,
                  order = 2,
               },
               buttons_x_row = {
                  type = 'range',
                  name = L['Max buttons x row'],
                  get = get_opt,
                  set = set_opt,
                  min = 4,
                  max = 30,
                  step = 1,
                  bigStep = 1,
                  isPercent = false,
                  order = 3,
               },
               show_keys = {
                  type = 'toggle',
                  name = L['Show keys'],
                  get = get_opt,
                  set = function(info, value)
                     db.general.show_keys = value
                     if (not value) and famBags.Buttons[KEYRING_CONTAINER] then
                        for _, b in ipairs(famBags.Buttons[KEYRING_CONTAINER]) do
                           b:Hide()
                        end
                     end
                     famBags:BuildBagIndexes()
                     set_opt(info, value)
                  end,
                  order = 4,
               },
            },
            order = 1,
         },
         sections = {
            name = L['Sections layout'],
            type = 'group',
            inline = true,
            args = {
               hpadding = {
                  type = 'range',
                  name = L['Horizontal padding'],
                  get = get_opt,
                  set = set_opt,
                  min = 0,
                  max = 30,
                  step = 0.1,
                  bigStep = 0.1,
                  isPercent = false,
                  order = 1,
               },
               vpadding = {
                  type = 'range',
                  name = L['Vertical padding'],
                  get = get_opt,
                  set = set_opt,
                  min = 0,
                  max = 30,
                  step = 0.1,
                  bigStep = 0.1,
                  isPercent = false,
                  order = 2,
               },
               hspacing = {
                  type = 'range',
                  name = L['Horizontal spacing'],
                  get = get_opt,
                  set = set_opt,
                  min = 0,
                  max = 30,
                  step = 0.1,
                  bigStep = 0.1,
                  isPercent = false,
                  order = 3,
               },
               vspacing = {
                  type = 'range',
                  name = L['Vertical spacing'],
                  get = get_opt,
                  set = set_opt,
                  min = 0,
                  max = 30,
                  step = 0.1,
                  bigStep = 0.1,
                  isPercent = false,
                  order = 4,
               },
            },
            order = 2,
         },
         buttons = {
            name = L['Items layout'],
            type = 'group',
            inline = true,
            args = {
               hpadding = {
                  type = 'range',
                  name = L['Horizontal padding'],
                  get = get_opt,
                  set = set_opt,
                  min = 0,
                  max = 30,
                  step = 0.1,
                  bigStep = 0.1,
                  isPercent = false,
                  order = 1,
               },
               vpadding = {
                  type = 'range',
                  name = L['Vertical padding'],
                  get = get_opt,
                  set = set_opt,
                  min = 0,
                  max = 30,
                  step = 0.1,
                  bigStep = 0.1,
                  isPercent = false,
                  order = 2,
               },
               hspacing = {
                  type = 'range',
                  name = L['Horizontal spacing'],
                  get = get_opt,
                  set = set_opt,
                  min = 0,
                  max = 30,
                  step = 0.1,
                  bigStep = 0.1,
                  isPercent = false,
                  order = 3,
               },
               vspacing = {
                  type = 'range',
                  name = L['Vertical spacing'],
                  get = get_opt,
                  set = set_opt,
                  min = 0,
                  max = 30,
                  step = 0.1,
                  bigStep = 0.1,
                  isPercent = false,
                  order = 4,
               },
               h2 = {
                  type = 'header',
                  name = '',
                  order = 5,
               },
               size = {
                  type = 'range',
                  name = L['Size'],
                  get = get_opt,
                  set = function(info, value)
                     set_opt(info, value)
                     famBags.Buttons:UpdateButtonsScale()
                  end,
                  min = 0.1,
                  max = 2,
                  step = 0.05,
                  bigStep = 0.05,
                  isPercent = true,
                  order = 6,
               },
               quality_glow = {
                  type = 'select',
                  name = L['Quality glow'],
                  get = get_opt,
                  set = function(info, value)
                     set_opt(info, value)
                     famBags.Buttons:UpdateAllButtonsQuality()
                  end,
                  values = function(info)
                     return quality_glows
                  end,
                  order = 7,
               },
               quality_glow_alpha = {
                  type = 'range',
                  name = L['Quality glow intensity'],
                  get = get_opt,
                  set = function(info, value)
                     set_opt(info, value)
                     famBags.Buttons:UpdateAllButtonsQuality()
                  end,
                  min = 0,
                  max = 1,
                  step = 0.05,
                  bigStep = 0.05,
                  isPercent = true,
                  hidden = function() return db.buttons.quality_glow == 'none' end,
                  order = 8,
               },
               quality_glow_inset = {
                  type = 'range',
                  name = L['Quality glow inset'],
                  get = get_opt,
                  set = function(info, value)
                     set_opt(info, value)
                     famBags.Buttons:UpdateAllButtonsQuality()
                  end,
                  min = -10,
                  max = 10,
                  step = 1,
                  bigStep = 1,
                  isPercent = false,
                  hidden = function() return db.buttons.quality_glow == 'none' end,
                  order = 9,
               },
               h3 = {
                  type = 'header',
                  name = '',
                  order = 10,
               },
               indicators_enabled = {
                  type = 'toggle',
                  name = L['Enable indicators'],
                  get = get_opt,
                  set = function(info, value)
                     db.buttons.indicators_enabled = value
                     famBags.Buttons:UpdateAllIndicators()
                     famBags:LayoutBagsButtons()
                  end,
                  order = 11,
               },
               indicators_size = {
                  type = 'range',
                  name = L['Indicators size'],
                  get = get_opt,
                  set = function(info, value)
                     db.buttons.indicators_size = value
                     famBags.Buttons:UpdateAllIndicators()
                     famBags:LayoutBagsButtons()
                  end,
                  min = 0.1,
                  max = 2,
                  step = 0.01,
                  bigStep = 0.01,
                  isPercent = true,
                  hidden = function() return not db.buttons.indicators_enabled end,
                  order = 12,
               },
               indicators_offset = {
                  type = 'range',
                  name = L['Indicators offset'],
                  get = get_opt,
                  set = function(info, value)
                     db.buttons.indicators_offset = value
                     famBags.Buttons:UpdateAllIndicators()
                     famBags:LayoutBagsButtons()
                  end,
                  min = -10,
                  max = 10,
                  step = 1,
                  bigStep = 1,
                  isPercent = false,
                  hidden = function() return not db.buttons.indicators_enabled end,
                  order = 13,
               },
            },
            order = 3,
         },
         counts = self:GetTextOptions(nil, nil, db.counts, false, true)
      },
   }
end

function famBagsOptions:GetMainWindowOptions()
   return {
      name = L['Main bags'],
      desc = L['Main bags options'],
      type = 'group',
      args = {
         appearance = {
            name = L['Appearance'],
            type = 'group',
            args =  self:GetAppearanceOptions({famBags.famBagsFrame.bodyFrame}, self.db.profile.mainwindow)
         },
         title = {
            name = L['Title'],
            type = 'group',
            args =  self:GetTitleOptions({famBags.famBagsFrame}, self.db.profile.mainwindow, true)
         }
      },
   }
end

function famBagsOptions:GetMoneyOptions()
   local db = self.db.profile.money
   local function get_opt(info)
      local id = info[#info]
      return db[id]
   end
   local function set_opt(info, value)
      local id = info[#info]
      db[id] = value
      famBags:UpdateMoneyFrame()
      famBags:UpdateCurrency()
   end
   local args = self:GetAppearanceOptions({famBags.famBagsMoneyFrame}, db)
   args['text'] = self:GetTextOptions({famBags.famBagsMoneyFrame}, 'moneyText', db.text)
   args['main'] = {
      name = L['General'],
      type = 'group',
      inline = true,
      args = {
         enabled = {
            type = 'toggle',
            name = L['Enabled'],
            get = get_opt,
            set = set_opt,
            order = 1,
         },
         show_silver = {
            type = 'toggle',
            name = '|TInterface\\MoneyFrame\\UI-SilverIcon:16:16:3:0:64:64:4:60:4:60|t ',
            desc = L['Show silver'],
            width = 'half',
            get = get_opt,
            set = set_opt,
            order = 2,
         },
         show_copper = {
            type = 'toggle',
            name = '|TInterface\\MoneyFrame\\UI-CopperIcon:16:16:3:0:64:64:4:60:4:60|t ',
            desc = L['Show copper'],
            width = 'half',
            get = get_opt,
            set = set_opt,
            order = 3,
         },
      },
      order = 1,
   }
   return {
      name = L['Money'],
      desc = L['Money options'],
      type = 'group',
      args = args,
   }
end

function famBagsOptions:GetBankWindowOptions()
   local db = self.db.profile.bankwindow
   local function get_opt(info)
      local id = info[#info]
      return db[id]
   end
   local function set_opt(info, value)
      local id = info[#info]
      db[id] = value
      famBags:UpdateBlizzardBankFrame()
   end
   local args = self:GetAppearanceOptions({famBags.famBagsBlizzardBankFrame}, db)
   args['main'] = {
      name = L['General'],
      type = 'group',
      inline = true,
      args = {
         enabled = {
            type = 'toggle',
            name = L['Override bank'],
            name = L['Override Blizzard bank main window. Requires restart.'],
            get = get_opt,
            set = set_opt,
            order = 1,
         },
      },
      order = 1,
   }
   return {
      name = L['Blizzard bank'],
      desc = L['Blizzard bank window override options'],
      type = 'group',
      args = args,
   }
end

function famBagsOptions:GetBagsWindowOptions()
   return {
      name = L['Bag slots'],
      desc = L['Bag slots options'],
      type = 'group',
      args = self:GetAppearanceOptions({famBags.famBagsBagSlotsFrame}, self.db.profile.bagswindow),
   }
end

function famBagsOptions:GetBankOptions()
   return {
      name = L['Bank'],
      desc = L['Bank options'],
      type = 'group',
      args = {
         appearance = {
            name = L['Appearance'],
            type = 'group',
            args =  self:GetAppearanceOptions({famBags.famBankFrame.bodyFrame}, self.db.profile.bank)
         },
         title = {
            name = L['Title'],
            type = 'group',
            args =  self:GetTitleOptions({famBags.famBankFrame}, self.db.profile.bank, true)
         }
      },
   }
end

function famBagsOptions:GetPresetsOptions()
   local current_preset
   return {
      name = L['Presets'],
      desc = L['Presets options'],
      type = 'group',
      args = {
         presets = {
            type = 'select',
            name = L['Presets'],
            get = function(info,key)
               if not famBags.PresetNames[current_preset] then
                  current_preset = 1
               end
               return current_preset
            end,
            set = function(info, value)
               current_preset = value
            end,
            values = function(info)
               return famBags.PresetNames
            end,
            order = 1,
         },
         apply = {
            type = 'execute',
            name = L['Apply preset'],
            func = function(info)
               wipe(famBags.db.profile)
               local defaults = famBags:GetDBDefaults().profile
               for k, t in pairs(defaults) do
                  famBags.db.profile[k] = CopyTable(t)
               end
               copy_data(famBags.Presets[current_preset], famBags.db.profile)
               self:FullRefresh()
            end,
            order = 2,
         }
      },
   }
end

function famBagsOptions:GetSectionsOptions()
   if not rawget(SECTIONS, CURRENT_SECTION) then
      CURRENT_SECTION = next(SECTIONS)
      CURRENT_SECTION_DB = self.db.profile.sections[CURRENT_SECTION]
   end
   local copy_from

   local function get_sections_frames(frameIndex)
      frames = {}
      local sections_type = self.db.profile.layout.general.sections
      for bag in pairs(famBags.BagIndexes) do
         local s = sections_type == 'bags' and famBags.Sections[bag] or famBags.Sections[bag].categories
         if s and s[CURRENT_SECTION] then
            if sections_type == 'bags' then
               for _, sections in pairs(s[CURRENT_SECTION]) do
                  if frameIndex and sections[frameIndex] then
                     table.insert(frames, sections[frameIndex])
                  else
                     table.insert(frames, sections)
                  end
               end
            elseif sections_type == 'categories' then
               if s[CURRENT_SECTION] then
                  if frameIndex and s[CURRENT_SECTION][frameIndex] then
                     table.insert(frames, s[CURRENT_SECTION][frameIndex])
                  else
                     table.insert(frames, s[CURRENT_SECTION])
                  end
               end
            end
         end
      end
      return frames
   end
   SECTION_ARGS = {
      sections = {
         type = 'select',
         name = L['Type'],
         get = function(info,key)
            if not rawget(SECTIONS, CURRENT_SECTION) then
               CURRENT_SECTION = next(SECTIONS)
            end
            return CURRENT_SECTION
         end,
         set = function(info, value)
            CURRENT_SECTION = value
            CURRENT_SECTION_DB = self.db.profile.sections[CURRENT_SECTION]
            local frames = get_sections_frames('bodyFrame')
            SECTION_ARGS.appearance.args = self:GetAppearanceOptions(frames, CURRENT_SECTION_DB)
            frames = get_sections_frames()
            SECTION_ARGS.title.args = self:GetTitleOptions(frames, CURRENT_SECTION_DB, true)
            local acr = LibStub('AceConfigRegistry-3.0')
            acr:NotifyChange('famBags Sections')
         end,
         values = function(info)
            return SECTIONS
         end,
         order = 1,
      },
      s1 = {
         type = 'description',
         name = '',
         order = 2,
         width = 'normal',
      },
      copy_from = {
         type = 'select',
         name = L['Copy from'],
         get = function(info,key)
            if not rawget(SECTIONS, copy_from) then
               copy_from = next(SECTIONS)
            end
            return copy_from
         end,
         set = function(info, value)
            copy_from = value
         end,
         values = function(info)
            return SECTIONS
         end,
         order = 3,
      },
      copy = {
         type = 'execute',
         name = L['Copy'],
         width = 'half',
         func = function(info)
            copy_section_db(copy_from)
            local frames = get_sections_frames('bodyFrame')
            for _, frame in ipairs(frames) do
               if (frame) then
                  local parent = frame:GetParent()
                  famBags:SetFrameBackdrop(frame, CURRENT_SECTION_DB)
                  famBags:SetFrameBackdrop(parent.titleFrame, CURRENT_SECTION_DB.title)
                  famBags:SetTitleInset(parent, CURRENT_SECTION_DB.title.is_inset, CURRENT_SECTION_DB.title.inset, CURRENT_SECTION_DB.title.text.padding)
                  famBags:SetTextLayout(parent.titleFrame, 'titleText', CURRENT_SECTION_DB.title.text, nil)
               end
            end
            famBags:LayoutBagsButtons()
         end,
         disabled = function() 
            return CURRENT_SECTION == copy_from
         end,
         order = 4,
      },
      h2 = {
         type = 'header',
         name = '',
         order = 5,
      },
      appearance = {
         name = L['Appearance'],
         type = 'group',
         args =  self:GetAppearanceOptions(get_sections_frames('bodyFrame'), CURRENT_SECTION_DB)
      },
      title = {
         name = L['Title'],
         type = 'group',
         args =  self:GetTitleOptions(get_sections_frames(), CURRENT_SECTION_DB, true)
      }
   }

   return {
      name = L['Sections'],
      desc = L['Sections options'],
      type = 'group',
      args = SECTION_ARGS
   }
end

function famBagsOptions:GetAppearanceOptions(frames, db)
   local function get_opt(info)
      local id = info[#info]
      return db.bg[id]
   end
   local function set_opt(info, value)
      local id = info[#info]
      db.bg[id] = value
      if frames then
         for _, frame in ipairs(frames) do
            if (frame) then
               famBags:SetFrameBackdrop(frame, db)
            end
         end
      end
   end
   return {
      bg = {
         name = L['Background'],
         type = 'group',
         inline = true,
         args = {
            type = {
               type = 'select',
               dialogControl = MULTISELECT_CONTROL,
               name = L['Type'],
               get = get_opt,
               set = set_opt,
               values = function(info)
                  return bg_types
               end,
               order = 1,
            },
            h2 = {
               type = 'header',
               name = '',
               order = 2,
            },
            texture = {
               type = 'select',
               dialogControl = AceGUI.WidgetRegistry['LSM30_Background'] and 'LSM30_Background' or nil,
               name = L['Texture'],
               get = get_opt,
               set = set_opt,
               values = function(info)
                  return LSM:HashTable('background')
               end,
               order = 3,
               hidden = function(info)
                  return db.bg.type ~= 4 --[[ Texture--]]
               end,
            },
            color1 = {
               type = 'color',
               width = 'half',
               name = function()
                  if db.bg.type == 1 --[[ Solid--]] then
                     return L['Color']
                  else
                     return L['Start color']
                  end
               end,
               hasAlpha = true,
               get = function(info)
                  return unpack(get_opt(info))
               end,
               set = function(info, r, g, b, a)
                  set_opt(info, {r, g, b, a})
               end,
               disabled = function(info)
                  return false
               end,
               order = 4,
            },
            color2 = {
               type = 'color',
               name = L['End color'],
               hasAlpha = true,
               width = 'half',
               get = function(info)
                  return unpack(get_opt(info))
               end,
               set = function(info, r, g, b, a)
                  set_opt(info, {r, g, b, a})
               end,
               disabled = function(info)
                  return false
               end,
               hidden = function(info)
                  return db.bg.type == 1 --[[ Solid--]] or db.bg.type == 4 --[[ Texture--]]
               end,
               order = 5,
            },
            inset = {
               type = 'range',
               name = L['Inset'],
               get = get_opt,
               set = set_opt,
               min = -50,
               max = 50,
               step = 1,
               bigStep = 1,
               isPercent = false,
               order = 6,
            },
         }
      },
      br = {
         name = L['Border'],
         type = 'group',
         inline = true,
         args = {
            br_texture = {
               type = 'select',
               dialogControl = AceGUI.WidgetRegistry['LSM30_Border'] and 'LSM30_Border' or nil,
               name = L['Texture'],
               get = get_opt,
               set = set_opt,
               values = function(info)
                  return LSM:HashTable('border')
               end,
               order = 2,
            },
            br_color = {
               type = 'color',
               name = L['Color'],
               hasAlpha = true,
               width = 'half',
               get = function(info)
                  return unpack(get_opt(info))
               end,
               set = function(info, r, g, b, a)
                  set_opt(info, {r, g, b, a})
               end,
               disabled = function(info)
                  return db.bg.br_texture == 'None'
               end,
               order = 3,
            },
            br_size = {
               type = 'range',
               name = L['Edge size'],
               get = get_opt,
               set = set_opt,
               min = 0,
               max = 50,
               step = 1,
               bigStep = 1,
               isPercent = false,
               disabled = function(info)
                  return db.bg.br_texture == 'None'
               end,
               order = 4,
            },
         }
      }

   }
end

function famBagsOptions:GetTextOptions(frames, name, db, refreshLayout, counts)
   local function get_opt(info)
      local id = info[#info]
      return db[id]
   end
   local function set_opt(info, value)
      local id = info[#info]
      db[id] = value
      if counts then
         for _, indexes in pairs(famBags.BagIndexes) do
            for _, i in ipairs(indexes) do
               if famBags.Buttons[i] then
                  for _, b in ipairs(famBags.Buttons[i]) do
                     famBags:SetTextLayout(b, _G[b:GetName().."Count"], db, nil)
                  end
               end
            end
         end
      elseif frames then
         for _, frame in ipairs(frames) do
            if frame then
               famBags:SetTextLayout(frame, name, db, nil)
            end
         end
      end
      if refreshLayout then
         famBags:LayoutBagsButtons()
      end
   end
   return {
      name = (counts) and L['Counts'] or L['Text'],
      type = 'group',
      inline = true,
      args = {
         font = {
            type = 'select',
            dialogControl = AceGUI.WidgetRegistry['LSM30_Font'] and 'LSM30_Font' or nil,
            name = L['Font'],
            get = get_opt,
            set = set_opt,
            values = function(info)
               return LSM:HashTable('font')
            end,
            order = 1,
         },
         flags = {
            type = 'multiselect',
            name = L['Flags'],
            dialogControl = 'Dropdown',
            get = function(info, key)
               return db.flags[key]
            end,
            set = function(info, key, value)
               db.flags[key] = value
               if frames then
                  for _, frame in ipairs(frames) do
                     if frame then
                        famBags:SetTextLayout(frame, name, db)
                     end
                  end
               end
            end,
            values = function(info)
               return text_flags
            end,
            order = 2,
         },
         size = {
            type = 'range',
            name = L['Size'],
            get = get_opt,
            set = set_opt,
            min = 4,
            max = 20,
            step = 1,
            bigStep = 1,
            isPercent = false,
            order = 3,
         },
         color = {
            type = 'color',
            name = L['Color'],
            hasAlpha = false,
            width = 'half',
            get = function(info)
               return unpack(get_opt(info))
            end,
            set = function(info, r, g, b)
               set_opt(info, {r, g, b})
            end,
            disabled = function(info)
               return false
            end,
            order = 4,
         },
         halign = {
            type = 'select',
            name = L['Horizontal align'],
            get = get_opt,
            set = set_opt,
            values = function(info)
               return horizontal_align
            end,
            order = 5,
         },
         valign = {
            type = 'select',
            name = L['Vertical align'],
            get = get_opt,
            set = set_opt,
            values = function(info)
               return vertical_align
            end,
            order = 6,
         },
         padding = {
            type = 'range',
            name = L['Padding'],
            get = get_opt,
            set = set_opt,
            min = 0,
            max = 20,
            step = 1,
            bigStep = 1,
            isPercent = false,
            order = 7,
         },
      }
   }
end

function famBagsOptions:GetTitleOptions(frames, db, refreshLayout)
   local function get_opt(info)
      local id = info[#info]
      return db.title[id]
   end
   local function set_opt(info, value)
      local id = info[#info]
      db.title[id] = value
      if frames then
         for _, frame in ipairs(frames) do
            if frame then
               famBags:SetTitleInset(frame, db.title.is_inset, db.title.inset, db.title.text.padding)
            end
         end
      end
      if refreshLayout then
         --famBags.Sections:LayoutBags(self.db.profile.layout.general.sections)
         famBags:LayoutBagsButtons()
      end
   end
   local titleframes = {}
   if frames then
      for _, frame in ipairs(frames) do
         if frame and frame.titleFrame then
            table.insert(titleframes, frame.titleFrame)
         end
      end
   end
   local args = self:GetAppearanceOptions(titleframes, db.title)
   args['title'] = {
      name = L['General'],
      type = 'group',
      inline = true,
      args = {
         is_inset = {
            type = 'toggle',
            name = L['Is inset'],
            get = get_opt,
            set = set_opt,
            order = 1,
         },
         inset = {
            type = 'range',
            name = L['Inset'],
            get = get_opt,
            set = set_opt,
            min = 0,
            max = 50,
            step = 1,
            bigStep = 1,
            isPercent = false,
            hidden = function(info)
               return not db.title.is_inset
            end,
            order = 2,
         },
      },
      order = 1,
   }
   args['text'] = self:GetTextOptions(titleframes, 'titleText', db.title.text, refreshLayout)
   return args
end

function famBagsOptions:GetCategoryOperatorOptions(cat, t, order)
   cat = get_actual_category_name(cat)
   local db = self.db.profile.categories[cat]
   return {
      type = 'select',
      dialogControl = 'Dropdown',
      name = ' ',
      width = 'half',
      get = function(info, key)
         return db.Criterias[get_criteria_index(cat, t, 'Criterias')].operator or 'AND'
      end,
      set = function(info, key, value)
         db.Criterias[get_criteria_index(cat, t, 'Criterias')].operator = key
         famBags:LayoutBagsButtons()
      end,
      values = function(info)
         return logical_operators
      end,
      order = order
   }
end

function famBagsOptions:GetCategoryTypeOptions(cat, t, order, set)
   set = set or 'Types'
   cat = get_actual_category_name(cat)
   local db = get_type_db(cat, t, set)
   return {
      type = 'multiselect',
      dialogControl = 'Dropdown',
      name = t,
      width = 'double',
      get = function(info, key)
         return db[key]
      end,
      set = function(info, key, value)
         local st = get_subtypes(t, set)
         if value == false then
            value = nil
         elseif key == L['.: All :.'] then
            db[L['.: All :.']] = L['.: All :.']
            for n in pairs(st) do
               if n ~= L['.: All :.'] then
                  db[n] = nil
               end
            end
            famBags:LayoutBagsButtons()
            return
         end
         db[key] = value
         local found
         for n in pairs(st) do
            if n ~= L['.: All :.'] and db[n] then
               db[L['.: All :.']] = nil
               found = true
               break
            end
         end
         if not found then
            db[L['.: All :.']] = L['.: All :.']
         end
         famBags:LayoutBagsButtons()
      end,
      values = function(info)
         return get_subtypes(t, set)
      end,
      order = order
   }
end

function famBagsOptions:GetCategoryTypeDelete(cat, t, order, set)
   set = set or 'Types'
   return {
      type = 'execute',
      name = L['Delete'],
      width = 'half',
      func = function(info)
         delete_type(cat, t, set)
         famBags:LayoutBagsButtons()
      end,
      order = order
   }
end

function famBagsOptions:GetCategoryTypeSeparator(order, set)
   set = set or 'Types'
   return {
      type = 'header',
      name = '',
      width = 'half',
      order = order
   }
end

function famBagsOptions:GetCategoryTypeUp(cat, t, order)
   return {
		type = "execute",
		name = L['Up'],
		width = 'half',
		func = function() 
		   move_criteria(cat, t, -1) 
         famBags:LayoutBagsButtons()
		end,
		order = order, 
	}
end

function famBagsOptions:GetCategoryTypeDown(cat, t, order)
   return {
		type = "execute",
		name = L['Down'],
		width = 'half',
		func = function() 
		   move_criteria(cat, t, 1) 
         famBags:LayoutBagsButtons()
		end,
		order = order, 
	}
end

function famBagsOptions:GetCategoryTypesOptions(category, set)
   set = set or 'Types'
   category = get_actual_category_name(category)
   local db = get_type_db(category, nil, set)
   local args = {}
   local order = 1
   if db then
      for t, crit in pairs(db) do
         if set == 'Criterias' then
            t = crit.name
            args[t..'_oper'] = self:GetCategoryOperatorOptions(category, t, order)
            order = order + 1
         end
         args[t] = self:GetCategoryTypeOptions(category, t, order, set)
         if set == 'Criterias' then
            args[t..'_up'] = self:GetCategoryTypeUp(category, t, order + 1)
            args[t..'_down'] = self:GetCategoryTypeDown(category, t, order + 2)
            order = order + 2
         end
         args[t..'_del'] = self:GetCategoryTypeDelete(category, t, order + 1, set)
         args[t..'_sep'] = self:GetCategoryTypeSeparator(order + 2, set)
         order = order + 3
      end
   end
   return args
end

function famBagsOptions:GetCategoryOptions(category)
   local db = self.db.profile.categories[category]
   local current_type = ' '
   local current_criteria = ' '
   return {
      name = category,
      type = 'group',
      args = {
         up = {
            type = 'execute',
            name = L['Up'],
            width = 'half',
            func = function(info)
               move_category_up(category)
               famBags:LayoutBagsButtons()
            end,
            order = 1,
         },
         break1 = {
            type = 'description',
            name = '',
            order = 2,
         },
         down = {
            type = 'execute',
            name = L['Down'],
            width = 'half',
            func = function(info)
               move_category_down(category)
               famBags:LayoutBagsButtons()
            end,
            order = 3,
         },
         spacer1 = {
            type = 'description',
            name = '',
            desc = '',
            width = 'half',
            order = 4,
         },
         delete = {
            type = 'execute',
            name = L['Delete'],
            confirm = function(info)
               return L['Are you sure you want to delete selected category? This is irreversible.']
            end,
            func = function(info)
               delete_category(category)
               famBags:LayoutBagsButtons()
            end,
            order = 5,
         },
         header1 = {
            type = 'header',
            name = '',
            desc = '',
            order = 6,
         },
         types = {
            type = 'select',
            dialogControl = 'Dropdown',
            name = L['Types'],
            get = function(info, key)
               return current_type
            end,
            set = function(info, value)
               current_type = value
            end,
            values = function(info)
               return get_types()
            end,
            order = 7,
         },
         add_type = {
            type = 'execute',
            name = L['Add'],
            width = 'half',
            func = function(info)
               add_type(category, current_type)
               famBags:LayoutBagsButtons()
            end,
            disabled = function() return current_type == ' ' or (db.Types and db.Types[current_type]) end,
            order = 8,
         },
         spacer2 = {
            type = 'description',
            name = '',
            desc = '',
            width = 'half',
            order = 9,
         },
         criterias = {
            type = 'select',
            dialogControl = 'Dropdown',
            name = L['Other criterias'],
            get = function(info, key)
               return current_criteria
            end,
            set = function(info, value)
               current_criteria = value
            end,
            values = function(info)
               return get_criterias()
            end,
            order = 10,
         },
         add_criteria = {
            type = 'execute',
            name = L['Add'],
            width = 'half',
            func = function(info)
               add_criteria(category, current_criteria)
               famBags:LayoutBagsButtons()
            end,
            disabled = function() 
               return current_criteria == ' ' or (db.Criterias and get_criteria_index(category, current_criteria, 'Criterias')) 
            end,
            order = 11,
         },
         types_set = {
            name = L['Types set'],
            type = 'group',
            inline = true,
            args = self:GetCategoryTypesOptions(category),
            order = 12,
         },
         criterias_set = {
            name = L['Other criterias'],
            type = 'group',
            inline = true,
            args = self:GetCategoryTypesOptions(category, 'Criterias'),
            order = 13,
         }
      },
   }
end

function famBagsOptions:GetCategoriesOptions()
   local db = self.db.profile.categories

   local function category_exists(category)
      for k in pairs(db) do
         if strupper(strsub(k, 4)) == category then
            return true
         end
      end
      return false
   end
   local function get_new_category_name(name)
      return string.format('%02d', get_array_len(famBags.db.profile.categories) + 1)..'.'..name
   end

   CATEGORIES_OPTIONS = {
      name = L['Categories'],
      type = 'group',
      childGroups = 'tree',
      args = {
         __new_cat__ = {
            type = 'input',
            name = L['Add new category'],
            get = function(info,key)
               return ''
            end,
            set = function(info,value)
               local c = get_new_category_name(value)
               db[c] = {}
               CATEGORIES_OPTIONS.args[c] = self:GetCategoryOptions(c)
               local dbs = self.db.profile.sections
               dbs[c] = {}
               dbs[c].bg = famBags:GetBGDefaults()
               dbs[c].title = famBags:GetTitleDefaults()
               SECTIONS[c] = c
               famBags:BuildSortedCategories()
            end,
            validate = function(info, value)
               value = strupper(strtrim(value))
               if not value or #value < 3 then
                  return L['Name too short.']
               elseif category_exists(value) then
                  return L['Category already exists.']
               else
                  return true
               end
            end,
            order = 1,
         },
      },
   }

   if next(db) then
      for c in pairs(db) do
         CATEGORIES_OPTIONS.args[c] = self:GetCategoryOptions(c)
         SECTIONS[c] = c
      end
   end

   return CATEGORIES_OPTIONS
end

